"""
djangogcal.observer


"""

from django.db.models import signals
from gdata.calendar import SendEventNotifications
from gdata.calendar.data import CalendarEventEntry
from gdata.calendar.client import CalendarClient

from models import CalendarEvent

class CalendarObserver(object):
    """
    
    """
    
    DEFAULT_FEED = '/calendar/feeds/default/private/full'
    
    def __init__(self, email, password, feed=DEFAULT_FEED, client=None):
        """
        Initialize an instance of the CalendarObserver class.
        """
        self.adapters = {}
        self.email = email
        self.password = password
        self.feed = feed
        self.client = client
    
    def observe(self, model, adapter):
        """
        Establishes a connection between the model and Google Calendar, using
        adapter to transform data.
        """
        self.adapters[model] = adapter
        signals.post_save.connect(self.on_update, sender=model,
                                  dispatch_uid="djangogcal post-save signal")
        signals.post_delete.connect(self.on_delete, sender=model,
                                    dispatch_uid="djangogcal post-delete signal")
    
    def observe_related(self, model, related, selector):
        """
        Updates the Google Calendar entry for model when the related model is
        changed or deleted.  Selector should be a function object which accepts
        an instance of related as a parameter and returns an instance of type
        model.
        """
        def on_related_update(**kwargs):
            self.update(model, selector(kwargs['instance']))
        signals.post_save.connect(on_related_update, sender=related, weak=False)
        signals.post_delete.connect(on_related_update, sender=related,
                                    weak=False)
    
    def on_update(self, **kwargs):
        """
        Called by Django's signal mechanism when an observed model is updated.
        """
        self.update(kwargs['sender'], kwargs['instance'])
    
    def on_delete(self, **kwargs):
        """
        Called by Django's signal mechanism when an observed model is deleted.
        """
        self.delete(kwargs['sender'], kwargs['instance'])
    
    def get_client(self):
        """
        Get an authenticated gdata.calendar.client.CalendarClient instance.
        """
        if self.client is None:
            self.client = CalendarClient(source='django-gcal')
            self.client.ClientLogin(self.email, self.password, self.client.source)
        return self.client
    
    def get_event(self, client, instance, feed=None):
        """
        Retrieves the specified event from Google Calendar, or returns None
        if the retrieval fails.
        """
        if feed is None:
            feed = self.feed
        event_id = CalendarEvent.objects.get_event_id(instance, feed)
        try:
            event = client.GetEventEntry(event_id)
        except Exception:
            event = None
        return event
    
    def update(self, sender, instance):
        """
        Update or create an entry in Google Calendar for the given instance
        of the sender model type.
        """
        adapter = self.adapters[sender]
        if adapter.can_save(instance):
            client = self.get_client()
            feed = adapter.get_feed_url(instance) or self.feed
            event = self.get_event(client, instance, feed) or CalendarEventEntry()
            adapter.get_event_data(instance).populate_event(event)
            if adapter.can_notify(instance):
                event.send_event_notifications = SendEventNotifications(
                    value='true')
            if event.GetEditLink():
                client.Update(event)
            else:
                new_event = client.InsertEvent(event, insert_uri=feed)
                CalendarEvent.objects.set_event_id(instance, feed,
                                                   new_event.get_edit_link().href)
    
    def delete(self, sender, instance):
        """
        Delete the entry in Google Calendar corresponding to the given instance
        of the sender model type.
        """
        adapter = self.adapters[sender]
        feed = adapter.get_feed_url(instance) or self.feed
        if adapter.can_delete(instance):
            client = self.get_client()
            event = self.get_event(client, instance, feed)
            if event:
                if adapter.can_notify(instance):
                    event.send_event_notifications = SendEventNotifications(
                        value='true')
                client.Delete(event)
        CalendarEvent.objects.delete_event_id(instance, feed)



"""
djangogcal.adapter


"""

from datetime import datetime

from atom.data import Content, Title
from gdata.calendar.data import When, CalendarWhere, EventWho
from django.utils.tzinfo import FixedOffset, LocalTimezone

DATE_FORMAT = '%Y-%m-%dT%H:%M:%S.000Z'

def format_datetime(date):
    """
    A utility method that converts the datetime to UTC serialized
    for Google Calendar.
    """
    local = date.replace(tzinfo=LocalTimezone(date))
    return local.astimezone(FixedOffset(0)).strftime(DATE_FORMAT)

class CalendarEventData(object):
    """
    A data-structure which converts Python data-types into Google Data API
    objects which can be transmitted to Google services.
    """
    
    def __init__(self, start, end, title="", where=None, who=None, content=""):
        """
        Instantiates a new instance of CalendarEventData.
        """
        self.start = start
        self.end = end
        self.title = title
        self.where = where or []
        self.who = who or []
        self.content = content
    
    def populate_event(self, event):
        """
        Populates the parameters of a Google Calendar event object.
        """
        event.when = [When(
            start=format_datetime(self.start),
            end=format_datetime(self.end)
        )]
        event.title = Title(text=self.title)
        event.where = [CalendarWhere(value_string=x) for x in self.where]
        event.who = [EventWho(email=x) for x in self.who]
        event.content = Content(text=self.content)

class RawCalendarEventData(object):
    """
    A data-structure which accepts Google Calendar data types, for users who
    need access to advanced fields.
    """
    
    def __init__(self, when, **kwargs):
        """
        Instantiates a new instance of RawCalendarEventData.
        """
        self.when = when
        self.kwargs = kwargs
    
    def populate_event(self, event):
        """
        Populates the parameters of a Google Calendar event object.
        """
        event.when = self.when
        for key in self.kwargs:
            setattr(event, key, self.kwargs[key])

class CalendarAdapter(object):
    """
    
    """
    
    def __init__(self):
        """
        Instantiates a new instance of CalendarAdapter.
        """
        pass
    
    def can_save(self, instance):
        """
        Should return a boolean indicating whether the object can be stored or
        updated in Google Calendar.
        """
        return True
    
    def can_delete(self, instance):
        """
        Should return a boolean indicating whether the object can be deleted
        from Google Calendar.
        """
        return True

    def can_notify(self, instance):
        """
        Should return a boolean indicating whether Google Calendar should send
        event change notifications.
        """
        return False
    
    def get_event_data(self, instance):
        """
        This method should be implemented by users, and must return an object
        conforming to the CalendarEventData protocol.
        """
        raise NotImplementedError()

    def get_feed_url(self, instance):
        """
        This method may be implemented by users, and should return a string to 
        be used to specify the feed for the event.
        """
        raise None
    



from distutils.core import setup
setup(
    name = "djangogcal",
    packages = ["djangogcal", ],
    version = "0.2.1",
    description = "A Django application allowing developers to synchronise instances of their models with Google Calendar.",
    author = "Incuna Ltd",
    author_email = "admin@incuna.com",
    url = "http://incuna.com/",
)



"""
djangogcal.models


"""

from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType
from django.db import models

class CalendarEventManager(models.Manager):
    """
    A custom manager for CalendarEvent models, containing utility methods for
    dealing with content-types framework.
    """
    
    def get_event_id(self, obj, feed_id):
        """
        Gets the Google Calendar event-id for a model, or returns None.
        """
        ct = ContentType.objects.get_for_model(obj)
        try:
            event = self.get(content_type=ct, object_id=obj.pk, feed_id=feed_id)
            event_id = event.event_id
        except models.ObjectDoesNotExist:
            event_id = None
        return event_id
    
    def set_event_id(self, obj, feed_id, event_id):
        """
        Sets the Google Calendar event-id for a model.
        """
        ct = ContentType.objects.get_for_model(obj)
        try:
            event = self.get(content_type=ct, object_id=obj.pk, feed_id=feed_id)
            event.event_id = event_id
        except models.ObjectDoesNotExist:
            event = CalendarEvent(content_type=ct, object_id=obj.pk,
                                  feed_id=feed_id, event_id=event_id)
        event.save()
    
    def delete_event_id(self, obj, feed_id):
        """
        Deletes the record containing the event-id for a model.
        """
        ct = ContentType.objects.get_for_model(obj)
        try:
            event = self.get(content_type=ct, object_id=obj.pk, feed_id=feed_id)
            event.delete()
        except models.ObjectDoesNotExist:
            pass

class CalendarEvent(models.Model):
    """
    
    """
    
    # django.contrib.contenttypes 'magic'
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    object = generic.GenericForeignKey()
    
    # google calendar event_id and feed_id
    event_id = models.CharField(max_length=255)
    feed_id = models.CharField(max_length=255)
    
    # custom manager
    objects = CalendarEventManager()
    
    def __unicode__(self):
        """ Returns the string representation of the CalendarEvent. """
        return u"%s: (%s, %s)" % (self.object, self.feed_id, self.event_id)
